import json
import logging
import os
import random
import threading
import time
from collections import defaultdict
from random import randint
from typing import Optional, Dict, List, Set, Iterator

from messageflux.iodevices.base import InputTransaction, InputDeviceManager, InputDevice, ReadResult, \
    InputDeviceException
from messageflux.iodevices.file_system.file_system_device_manager_base import FileSystemDeviceManagerBase
from messageflux.iodevices.file_system.file_system_serializer import FileSystemSerializerBase, \
    DefaultFileSystemSerializer
from messageflux.metadata_headers import MetadataHeaders
from messageflux.utils import get_random_id
from messageflux.utils.filesystem import atomic_move, AtomicMoveException


class FileSystemInputTransaction(InputTransaction):
    """
    represents an InputTransaction for filesystem
    """

    MAX_POISON_COUNT = 3  # TODO: get this from the user.
    _POISON_COUNTS_PER_FILE: Dict[str, int] = defaultdict(lambda: 0)
    _logger = logging.getLogger(__name__)

    def __init__(self, device: 'FileSystemInputDevice', org_path: str, tmp_path: str):
        """
        :param device: the device that returned this transaction
        :param org_path: the original path of the file we read
        :param tmp_path: the temp path of the file we read
        """
        super(FileSystemInputTransaction, self).__init__(device=device)
        self._org_path = org_path
        self._tmp_path = tmp_path
        self._device_manager = device.manager
        self._device_manager.transaction_log.add_transaction(self)

    @property
    def org_path(self) -> str:
        """
        the original path of the file in this transaction
        """
        return self._org_path

    @property
    def tmp_path(self) -> str:
        """
        the temp (current) path of the file in this transaction
        """
        return self._tmp_path

    @staticmethod
    def _get_poison_filename(org_filename: str) -> str:
        """
        gets a filename for poison messages

        :param org_filename: the original filename
        :return: filepath for the poison message
        """
        basename = os.path.basename(org_filename)
        dirname = os.path.dirname(org_filename)
        poison_folder = os.path.join(dirname, 'POISON')  # TODO: get this from the user
        os.makedirs(poison_folder, exist_ok=True)
        poison_filepath = os.path.join(poison_folder, get_random_id() + basename)
        return poison_filepath

    @staticmethod
    def _calc_lockfile_name(tmp_folder: str, org_path: str) -> str:
        """
        calculates a name for the lockfile

        :param tmp_folder: the tmp folder
        :param org_path: the original file path
        :return: the path for the lockfile
        """
        return os.path.join(tmp_folder, '{}.lockfile'.format(os.path.basename(org_path)))

    @staticmethod
    def read_file(device: 'FileSystemInputDevice',
                  org_path: str,
                  tmp_folder: str,
                  with_transaction: bool,
                  serializer: FileSystemSerializerBase) -> Optional[ReadResult]:
        """
        reads the file from org_path, and adds the file to transaction if it exists, or delete if it doesn't
        :param device: the input device for this transaction
        :param org_path: the original path to this file
        :param tmp_folder: the temporary folder to use while holding files in transaction
        :param with_transaction: if set to 'False' no transaction is made, and the message is acked on read
        :param serializer: the serializer to use to deserialize the file
        """

        tmp_path = os.path.join(tmp_folder, get_random_id())
        try:
            if not atomic_move(org_path,
                               tmp_path,
                               FileSystemInputTransaction._calc_lockfile_name(tmp_folder,
                                                                              org_path)):
                return None
        except AtomicMoveException:
            FileSystemInputTransaction._logger.exception(f'Atomic move could not move the file {org_path}')
            return None

        with open(tmp_path, 'rb') as f:
            message = serializer.deserialize(f)

        if with_transaction:
            return ReadResult(message=message,
                              transaction=FileSystemInputTransaction(device=device,
                                                                     org_path=org_path,
                                                                     tmp_path=tmp_path))
        else:
            try:
                os.remove(tmp_path)
            except FileNotFoundError:
                pass

            return ReadResult(message)

    def _commit(self):
        """
        commits the transaction
        """
        try:
            os.remove(self._tmp_path)
        except FileNotFoundError:
            pass
        self._POISON_COUNTS_PER_FILE.pop(self._org_path, None)
        self._device_manager.transaction_log.remove_transaction(self)

    @staticmethod
    def rollback_path(tmp_path: str, org_path: str, max_poison_count: int):
        """
        rolls back the transaction

        :param tmp_path: the temp (current) path of the file
        :param org_path: the original path of the file
        :param max_poison_count: the maximum time we allow the file to be rolled back
        """
        FileSystemInputTransaction._POISON_COUNTS_PER_FILE[org_path] += 1
        if FileSystemInputTransaction._POISON_COUNTS_PER_FILE[org_path] >= max_poison_count:
            dest = FileSystemInputTransaction._get_poison_filename(org_path)
            FileSystemInputTransaction._POISON_COUNTS_PER_FILE.pop(org_path, None)
        else:
            dest = org_path

        atomic_move(tmp_path, dest, FileSystemInputTransaction._calc_lockfile_name(os.path.dirname(tmp_path), dest))

    def _rollback(self):
        """
        rolls back the transaction
        """
        try:
            self.rollback_path(tmp_path=self._tmp_path,
                               org_path=self._org_path,
                               max_poison_count=self.MAX_POISON_COUNT)
            self._device_manager.transaction_log.remove_transaction(self)
        except AtomicMoveException:
            self._logger.exception(f"Could not rollback file:{self._org_path}")


class TransactionLog:
    """
    this class is used to persist transaction log into filesystem, so it can be rolled back on process termination
    """
    _logger = logging.getLogger(__name__)

    def __init__(self, filepath: str):
        """
        :param filepath: the full path of the file used to persist this log
        """
        self._filepath = filepath
        self._transactions: Dict[str, str] = {}
        self._transactions.update(self._load_file(filepath))

    @staticmethod
    def _load_file(filepath: str) -> Dict[str, str]:
        """
        loads the transaction log from file

        :param filepath: the path of the file to read
        """
        if os.path.exists(filepath):
            with open(filepath, 'r') as f:
                data = json.load(f)
            return data
        else:
            return {}

    @property
    def filepath(self):
        """
        the path to the file holding this transaction log
        """
        return self._filepath

    def add_transaction(self, transaction: FileSystemInputTransaction):
        """
        adds a transaction to the log

        :param transaction: the transaction to add
        """
        self._transactions[transaction.tmp_path] = transaction.org_path
        try:
            self.write_log()
        except Exception:
            self._logger.warning("Couldn't save transaction log", exc_info=True)

    def remove_transaction(self, transaction: FileSystemInputTransaction):
        """
        removes a transaction from the log

        :param transaction: the transaction to remove
        """
        self._transactions.pop(transaction.tmp_path, None)
        try:
            self.write_log()
        except Exception:
            self._logger.warning("Couldn't save transaction log", exc_info=True)

    def rollback_all(self):
        """
        rolls back all the transaction in the log
        """
        for tmp_path, org_path in self._transactions.items():
            try:
                FileSystemInputTransaction.rollback_path(tmp_path=tmp_path,
                                                         org_path=org_path,
                                                         max_poison_count=FileSystemInputTransaction.MAX_POISON_COUNT)
            except AtomicMoveException:
                self._logger.exception(f"Could not rollback file:{org_path}")
        self._transactions.clear()
        try:
            self.write_log()
        except Exception:
            self._logger.warning("Couldn't save transaction log", exc_info=True)

    def write_log(self):
        """
        writes the transaction log to the file
        """
        if not self._transactions:
            try:
                os.remove(self._filepath)
            except OSError:
                pass
        else:
            with open(self._filepath, 'w') as f:
                json.dump(self._transactions, f)


class FileSystemInputDevice(InputDevice['FileSystemInputDeviceManager']):
    """
    An InputDevice, that reads from folder
    """

    _MAX_BATCH_SIZE = 300
    _MIN_BATCH_SIZE = 8
    _SLEEP_BETWEEN_BATCHES = 1
    STAT_HEADER_NAME = "__STAT__"
    FILENAME_HEADER_NAME = MetadataHeaders.FILENAME

    def __init__(self,
                 manager: 'FileSystemInputDeviceManager',
                 name: str,
                 tmp_folder: str,
                 queues_folder: str,
                 fifo: bool = True,
                 min_file_age: int = 0,
                 serializer: Optional[FileSystemSerializerBase] = None):
        """
        ctor

        :param manager: the device manager which created this input device
        :param name: the name of the queue to read from
        :param tmp_folder: the folder to create temporary files in
        :param queues_folder: the base folder for all queues
        :param fifo: should we read the files sorted by time (Notice! when fifo=false, you might get file starvation)
        :param min_file_age: the minimum time in seconds since last modification to file, before we to try to read it...
        :param serializer: the serializer to use for reading messages from files
        """
        super(FileSystemInputDevice, self).__init__(manager, name)
        self.min_file_age = min_file_age
        self._sorted = fifo
        self._tmp_folder = tmp_folder
        self._input_folder = os.path.join(queues_folder, name)
        try:
            os.makedirs(self._tmp_folder, exist_ok=True)
            os.makedirs(self._input_folder, exist_ok=True)
        except Exception as e:
            raise InputDeviceException('Error creating input device') from e

        self._current_batch_size = self._MIN_BATCH_SIZE
        self._current_generator: Optional[Iterator[Optional[os.DirEntry]]] = None
        self._black_listed_files: Set[str] = set()
        self._logger = logging.getLogger(__name__)
        self._serializer = serializer or DefaultFileSystemSerializer()

    def _get_sorted_filenames(self) -> List[os.DirEntry]:
        """
        returns a list of direntries for files under input_folder, sorted by mtime

        :return: sorted list of direntries
        """

        tmplist = []
        for direntry in os.scandir(self._input_folder):
            try:
                if direntry.path in self._black_listed_files:
                    continue
                mtime = direntry.stat(follow_symlinks=False).st_mtime
            except Exception:
                continue
            tmplist.append((direntry, mtime))
        sorted_list = sorted(tmplist, key=lambda t: t[1])
        result = []
        for tup in sorted_list:
            result.append(tup[0])
        return result

    def _get_generator_files(self) -> Iterator[Optional[os.DirEntry]]:
        """
        returns a generator of direntries for files under input_folder, unsorted
        None when have no more direnties in the scandir

        :return: generator of direntries
        """

        while True:
            for direntry in os.scandir(self._input_folder):
                yield direntry
            yield None

    def _get_unsorted_filenames(self) -> List[os.DirEntry]:
        """
        returns a batch of generator direntries for files under input_folder, unsorted
        None when have no more direnties in the scandir

        :return: generator of direntries
        """
        if not self._current_generator:
            self._current_generator = self._get_generator_files()
        batch = []
        assert self._current_generator is not None
        for direntry in self._current_generator:
            if direntry is None:
                break
            try:
                if direntry.path in self._black_listed_files:
                    continue

                if self.min_file_age > 0:
                    file_age = time.time() - direntry.stat().st_mtime
                    if file_age < self.min_file_age:
                        continue
            except Exception:
                continue
            batch.append(direntry)
            if len(batch) >= self._current_batch_size:
                break

        self._current_batch_size = max(self._MIN_BATCH_SIZE,
                                       min(len(batch), self._current_batch_size))
        random.shuffle(batch)

        return batch

    def _increase_batch_size(self):
        """
        increases the batch size for concurrent reading
        """
        self._current_batch_size = min(self._current_batch_size * 2, self._MAX_BATCH_SIZE)

    def _decrease_batch_size(self):
        """
        decreases the batch size for concurrent reading
        """
        self._current_batch_size = max(self._MIN_BATCH_SIZE, int(self._current_batch_size / 2))

    def _read_file(self, direntry: os.DirEntry, with_transaction: bool = True) -> Optional[ReadResult]:
        """
        tries to read a single filename from queue

        :param direntry: the input file to read

        :return: None, None if the file can't be read or (BytesIO, dict) if it succeeded
        """
        if direntry.path in self._black_listed_files:
            return None

        if not direntry.is_file():
            self._black_listed_files.add(direntry.path)
            return None

        self._logger.debug(f"found input file {direntry.name} in directory {self._input_folder}")

        file_path = direntry.path
        return FileSystemInputTransaction.read_file(self,
                                                    org_path=file_path,
                                                    tmp_folder=self._tmp_folder,
                                                    with_transaction=with_transaction,
                                                    serializer=self._serializer)

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional[ReadResult]:
        """
        this method returns a message from the file system

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds, if the device doesn't have a message to return, it will return None

        :param with_transaction: 'True' if the device should read message within transaction,
        or 'False' if the message is automatically committed

        :return: a ReadResult object or None if no message was available.
        the device headers contains the filename, and stat struct for the file
        """
        try:
            deadline = 0.0
            if timeout is not None:
                deadline = time.perf_counter() + timeout
            if self._sorted:
                while True:
                    input_files = self._get_sorted_filenames()
                    for direntry in input_files:
                        try:
                            fs_metadata = {self.FILENAME_HEADER_NAME: direntry.name,
                                           self.STAT_HEADER_NAME: direntry.stat()}
                        except FileNotFoundError:  # file was deleted before we read it
                            continue
                        read_result = self._read_file(direntry, with_transaction=with_transaction)
                        if read_result is not None:
                            read_result.device_headers.update(fs_metadata)
                            return read_result
                        if timeout is not None and time.perf_counter() >= deadline:
                            break

                    if timeout is not None and time.perf_counter() >= deadline:
                        break
                    cancellation_token.wait(self._SLEEP_BETWEEN_BATCHES)
            else:
                while True:
                    got_file = False
                    for direntry in self._get_unsorted_filenames():
                        got_file = True
                        try:
                            fs_metadata = {self.FILENAME_HEADER_NAME: direntry.name,
                                           self.STAT_HEADER_NAME: direntry.stat()}
                        except FileNotFoundError:  # file was deleted before we read it
                            continue
                        read_result = self._read_file(direntry, with_transaction=with_transaction)
                        if read_result is not None:
                            self._decrease_batch_size()
                            read_result.device_headers.update(fs_metadata)
                            return read_result

                    # couldn't read any file from batch. try bigger batch next time
                    self._increase_batch_size()
                    if timeout is not None and time.perf_counter() >= deadline:
                        break
                    if not got_file:
                        cancellation_token.wait(self._SLEEP_BETWEEN_BATCHES)

            return None

        except Exception as e:
            raise InputDeviceException('Error reading product from device') from e


class FileSystemInputDeviceManager(FileSystemDeviceManagerBase, InputDeviceManager[FileSystemInputDevice]):
    """
    this is an input manager that creates file system input devices
    """

    def __init__(self,
                 root_folder: str,
                 queue_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_QUEUES_SUB_DIR,
                 tmp_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_TMPDIR_SUB_DIR,
                 bookkeeping_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_BOOKKEEPING_SUB_DIR,
                 serializer: Optional[FileSystemSerializerBase] = None,
                 fifo: bool = True,
                 min_input_file_age: int = 0,
                 transaction_log_save_interval: int = 10):
        """
        :param root_folder: the root folder to use for the manager
        :param queue_dir_name: the name of the subdirectory under root_folder that holds the queues
        :param tmp_dir_name: the name of the subdirectory under root_folder to use for temp files
        :param bookkeeping_dir_name: the name of the subdirectory under root_folder that holds the book-keeping data
        :param serializer: the serializer to use to write messages to files. None will use the default serializer
        :param fifo: should we read the files sorted by time (Notice! when fifo=false, you might get file starvation)
        :param min_input_file_age: the minimum time in seconds since last modification, before we to try to read it...
        :param transaction_log_save_interval: the interval in seconds to save the transaction log
        """
        super().__init__(root_folder=root_folder,
                         queue_dir_name=queue_dir_name,
                         tmp_dir_name=tmp_dir_name,
                         bookkeeping_dir_name=bookkeeping_dir_name,
                         serializer=serializer)
        self._fifo = fifo
        self._min_input_file_age = min_input_file_age
        transaction_log_filename = os.path.join(self.bookkeeping_folder, f'{self._unique_manager_id}.transactionlog')
        self._transaction_log = TransactionLog(transaction_log_filename)
        self._transaction_log_save_interval = transaction_log_save_interval
        self._should_stop: threading.Event = threading.Event()
        self._transaction_log_thread: Optional[threading.Thread] = None
        self._logger = logging.getLogger(__name__)

    def _do_transaction_log_thread(self):
        """
        a thread that looks for old transaction logs, and rolls them back
        """
        while not self._should_stop.is_set():
            try:
                self._transaction_log.write_log()
                now = time.time()
                for direntry in os.scandir(self.bookkeeping_folder):
                    try:
                        if not direntry.is_file():
                            continue
                        transaction_log_file = direntry.path
                        if not transaction_log_file.endswith(".transactionlog"):
                            continue
                        file_age = now - direntry.stat().st_mtime
                        if file_age < self._transaction_log_save_interval * 3:
                            # don't roll back transaction logs that are too new (might be a file of another pod/process)
                            continue

                        temp_transaction_log_file_name = f'{transaction_log_file}.rolling_back.{randint(0, 10000)}'
                        os.rename(src=transaction_log_file, dst=temp_transaction_log_file_name)
                    except FileNotFoundError:
                        continue  # Couldn't handle the file, because someone else got to it first

                    try:
                        transaction_log = TransactionLog(temp_transaction_log_file_name)
                        transaction_log.rollback_all()
                    except Exception:
                        self._logger.exception(f"Error rolling back {transaction_log_file} file. removing")
                    finally:
                        if os.path.exists(temp_transaction_log_file_name):
                            os.remove(temp_transaction_log_file_name)
            except Exception:
                self._logger.exception('Error in transaction_log thread')
            # sleep for at least a few seconds, so we won't spam
            self._should_stop.wait(self._transaction_log_save_interval)

    @property
    def transaction_log(self) -> TransactionLog:
        """
        the transaction log
        """
        return self._transaction_log

    def get_input_device(self, name: str) -> FileSystemInputDevice:
        """
        Returns an input device by name

        :param name: the name of the device to read from
        :return: an input device for 'device_name'
        """
        try:
            return FileSystemInputDevice(manager=self,
                                         name=name,
                                         tmp_folder=self._tmp_folder,
                                         queues_folder=self._queues_folder,
                                         fifo=self._fifo,
                                         min_file_age=self._min_input_file_age,
                                         serializer=self._serializer)
        except Exception as e:
            raise InputDeviceException("Error getting input device") from e

    def connect(self):
        """
        connects to the device manager
        """
        try:
            self._create_all_directories()
            self._should_stop.clear()
            self._transaction_log_thread = threading.Thread(target=self._do_transaction_log_thread, daemon=True)
            self._transaction_log_thread.start()
        except Exception as ex:
            raise InputDeviceException('Error connection to Device Manager') from ex

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._should_stop.set()
        self._transaction_log.rollback_all()



from .file_system_message_store import FileSystemMessageStore
from .message_store_base import MessageStoreBase
from .message_store_input_device import MessageStoreInputDeviceManagerWrapper
from .message_store_output_device import MessageStoreOutputDeviceManagerWrapper



from .transformer_input_device import TransformerInputDeviceManager, InputTransformerBase
from .transformer_output_device import TransformerOutputDeviceManager, OutputTransformerBase
from .zlib_transformer import ZLIBTransformer



import os
from typing import Optional

from messageflux.logging.bulk_rotating_handler_base import BulkRotatingHandlerBase


class BulkRotatingFileHandler(BulkRotatingHandlerBase):
    """
    Handler for logging to a set of files, which switches from one file
    to the next when the current file reaches a certain size or time.
    """

    def __init__(self,
                 live_log_path: str,
                 rotated_log_path: str,
                 bkp_log_path: Optional[str] = None,
                 max_records: int = 1000,
                 max_time: int = 60,
                 live_log_prefix: str = ''):
        """
        Open a log file and use it as the stream for logging.
        when max_records are written, or max_time has passed, a rollover occurs
        rollover copies the live log from live_log_path to rotated_log_path

        :param live_log_path: the path to write the live log to
        :param rotated_log_path: the path to write the rotated log to
        :param bkp_log_path: the path to write to rotated log to, if writing to rotated_log_path fails
        :param max_records: the maximum number of records to write before rotation
        :param max_time: the maximum time (in seconds) to wait before rotation
        :param live_log_prefix: the prefix for live log file
        """
        self._rotated_log_path = os.path.abspath(rotated_log_path)
        os.makedirs(self._rotated_log_path, exist_ok=True)

        super(BulkRotatingFileHandler, self).__init__(live_log_path=live_log_path,
                                                      bkp_log_path=bkp_log_path,
                                                      max_records=max_records,
                                                      max_time=max_time,
                                                      live_log_prefix=live_log_prefix)

    def _move_log_to_destination(self, src_file: str):
        """
        this moves the live log from a file, to its destination (the rotated log path)
        """

        dst_filename = os.path.join(self._rotated_log_path, self._get_unique_log_filename())
        self._safe_move(src_file, dst_filename)



import contextlib

from time import perf_counter

from messageflux.utils import KwargsException


class ShortCircuitException(KwargsException):
    """
    this exception is raised when the device is in short circuit state
    """
    pass


class ShortCircuitDeviceBase:
    """
    a base logic for short circuit devices
    """

    def __init__(self, short_circuit_fail_count: int, short_circuit_time: float):
        """

        :param short_circuit_fail_count: the consecutive number of failures that will trigger a short circuit
        :param short_circuit_time: the time in seconds that the device will be in short circuit state
        """
        self._short_circuit_fail_count = short_circuit_fail_count
        self._short_circuit_time = short_circuit_time
        self._short_circuit_end_time: float = -1.0
        self._consecutive_failures_count = 0

    @property
    def is_in_short_circuit_state(self) -> bool:
        """
        returns whether this device is still short circuited.

        :return: True if the device is short-circuited, False otherwise
        """
        return self._short_circuit_end_time >= 0 and perf_counter() < self._short_circuit_end_time

    @property
    def consecutive_failures_count(self) -> int:
        """
        returns the current consecutive failures count
        """
        return self._consecutive_failures_count

    def _validate_short_circuit(self):
        if self.is_in_short_circuit_state:
            raise ShortCircuitException("Device is in short circuited state")

        self._short_circuit_end_time = -1.0

    def _report_failure(self):
        self._consecutive_failures_count += 1
        if self._consecutive_failures_count >= self._short_circuit_fail_count:
            self._short_circuit_end_time = perf_counter() + self._short_circuit_time
            self._consecutive_failures_count = 0  # we are now in short circuit state, so reset the fail count

    def _report_success(self):
        self._consecutive_failures_count = 0

    @contextlib.contextmanager
    def _failure_count_context(self):
        try:
            yield self
        except Exception:
            self._report_failure()
            raise
        else:
            self._report_success()



import logging
import threading
from time import perf_counter
from typing import Optional, Collection, List, Callable

from messageflux import InputDevice, ReadResult
from messageflux.iodevices.base import InputDeviceManager, InputDeviceException


class CollectionInputDevice(InputDevice['CollectionInputDeviceManager']):
    """
    this class reads from a collection of input devices.

    it iterates over the collection, and returns the first successful item read from the device
    """
    _SLEEP_BETWEEN_ITERATIONS: float = 0.1

    def __init__(self,
                 device_manager: 'CollectionInputDeviceManager',
                 device_name: str,
                 input_devices: Collection[InputDevice]):
        """
        ctor

        :param device_name: the name of this device
        :param input_devices: the devices that this aggregate device reads from

        """
        super(CollectionInputDevice, self).__init__(device_manager, device_name)
        self._input_devices = input_devices
        self._logger = logging.getLogger(__name__)

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional['ReadResult']:
        """
        reads a message from device

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds, if the device doesn't have a message to return, it will return None

        :param with_transaction: 'True' if the device should read message within transaction,
        or 'False' if the message is automatically committed

        :return: a ReadResult object or None if no message was available.
        the device headers, can contain extra information about the device that returned the message
        """
        end_time = 0.0
        if timeout is not None:
            end_time = perf_counter() + timeout

        while True:
            failures = []
            for curr_device in self._input_devices:
                try:
                    read_result = curr_device._read_message(cancellation_token=cancellation_token,
                                                            timeout=0,
                                                            with_transaction=with_transaction)
                    if read_result is not None:
                        return read_result
                    if timeout is not None and perf_counter() >= end_time:
                        break
                except Exception as ex:
                    failures.append(ex)
                    self._logger.warning(f"Error reading from {curr_device.name} device", exc_info=True)
                    if timeout is not None and perf_counter() >= end_time:
                        break

            if len(failures) >= len(self._input_devices):
                self._logger.error('Error reading from CollectionInputDevice - all devices failed')
                raise InputDeviceException('Error reading from CollectionInputDevice - all devices failed',
                                           inner_exceptions=failures)

            if timeout is not None and perf_counter() >= end_time:
                return None
            else:
                cancellation_token.wait(self._SLEEP_BETWEEN_ITERATIONS)

    def close(self):
        """
        closes the connection to device
        """
        for device in self._input_devices:
            try:
                device.close()
            except Exception:
                self._logger.warning(f'Error closing underlying device {device.name}', exc_info=True)


class CollectionInputDeviceManager(InputDeviceManager[CollectionInputDevice]):
    """
    This class is used to create Collection InputDevices
    """

    def __init__(self,
                 inner_managers: List[InputDeviceManager],
                 collection_maker: Callable[[List[InputDevice]], Collection[InputDevice]]):
        """
        This class is used to create Collection InputDevices

        :param inner_managers: the actual InputDeviceManager instances to generate devices from
        :param collection_maker: the callable to make the iterable collection from list of devices
        """
        self._inner_managers = inner_managers
        self._logger = logging.getLogger(__name__)
        self._collection_maker = collection_maker

    def connect(self):
        """
        connects to device manager
        """
        failures = []
        for manager in self._inner_managers:
            try:
                manager.connect()
            except Exception as e:
                failures.append(e)
                self._logger.warning(
                    f'Error connecting to underlying manager: {type(manager).__name__}', exc_info=True)

        if len(failures) >= len(self._inner_managers):
            self._logger.error("Couldn't connect to any underlying manager")
            raise InputDeviceException('Error connecting to underlying device managers', inner_exceptions=failures)

    def disconnect(self):
        """
        disconnects from the device manager
        """
        for manager in self._inner_managers:
            try:
                manager.disconnect()
            except Exception:
                self._logger.warning(
                    f'Error closing underlying manager {type(manager).__name__}', exc_info=True)

    def get_input_device(self, name: str) -> CollectionInputDevice:
        """
        Returns an input device by name

        :param name: the name of the device to read from
        :return: an input device for 'device_name'
        """
        devices: List[InputDevice] = []
        failures = []
        for manager in self._inner_managers:
            try:
                device = manager.get_input_device(name)
                devices.append(device)
            except Exception as ex:
                failures.append(ex)
                self._logger.warning(
                    f"Error creating input device {name} from {type(manager).__name__} manager", exc_info=True)
        if not devices:
            self._logger.error("Couldn't create any input device")
            raise InputDeviceException(f"Couldn't create input device '{name}'", inner_exceptions=failures)

        return CollectionInputDevice(self, name, self._collection_maker(devices))



import threading
from abc import ABCMeta, abstractmethod
from enum import Enum, unique
from threading import Event
from typing import Optional, List, TYPE_CHECKING

from messageflux.utils import KwargsException

if TYPE_CHECKING:
    from messageflux.iodevices.base.input_devices import InputDevice, ReadResult


class WrongTransactionStateException(KwargsException):
    """
    this exception is raised when transaction is in the wrong state for operation
    """
    pass


@unique
class TransactionState(Enum):
    """
    represents the current state of the transaction
    """
    ACTIVE = 'ACTIVE'
    COMMITTED = 'COMMITTED'
    ROLLEDBACK = 'ROLLEDBACK'


class InputTransaction(metaclass=ABCMeta):
    """
    this is the base class for input transaction objects.

    after reading a message from the device (using 'with_transaction=True') the reader should commit/rollback
    the transaction, to signal the device if the message is done processing or not

    calling commit/rollback on a transaction, finishes it. calling commit or rollback multiple times on a transaction
    is legal but will only execute once. calling rollback after commit or vice versa, will raise an exception

    if using the transaction as context (i.e 'with transaction:') it will commit at the end of the context.
    if an exception was raised during the context, the transaction will be rolled back.
    """

    def __init__(self, device: 'InputDevice'):
        """
        :param device: the input device that returned that transaction
        """
        self._device: 'InputDevice' = device
        self._finished: Event = Event()
        self._state: TransactionState = TransactionState.ACTIVE

    @property
    def state(self) -> TransactionState:
        """
        the current state of the transaction
        """
        return self._state

    @property
    def device(self) -> 'InputDevice':
        """
        :return: the input device that returned that transaction
        """
        return self._device

    @property
    def finished(self) -> bool:
        """
        :return: 'True' if the transaction was committed/rolled-back, 'False' otherwise
        """
        return self._finished.is_set()

    def __enter__(self):
        if self.finished:
            raise WrongTransactionStateException('Cannot enter an already finished transaction')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.finished:
            return
        if (exc_type, exc_val, exc_tb) == (None, None, None):
            self.commit()  # no exception, commit the transaction
        else:
            self.rollback()  # exception, rollback the transaction

    def wait_for_finish(self, timeout: Optional[float] = None) -> bool:
        """
        waits for the transaction to finish (commit/rollback).
        if timeout is not None, after timeout seconds, it will return finished

        :param timeout: the timeout (in seconds) to wait for the transaction to finish

        :return: the value of finished
        """
        return self._finished.wait(timeout=timeout)

    def commit(self):
        """
        commits the transaction
        """
        if self.state == TransactionState.ROLLEDBACK:
            raise WrongTransactionStateException('Transaction Was Already Rolled Back')
        if self.state == TransactionState.COMMITTED:
            return

        self._commit()
        self._state = TransactionState.COMMITTED
        self._finished.set()

    def rollback(self):
        """
        rolls back the transaction
        """
        if self.state == TransactionState.COMMITTED:
            raise WrongTransactionStateException('Transaction Was Already Committed')
        if self.state == TransactionState.ROLLEDBACK:
            return

        self._rollback()
        self._state = TransactionState.ROLLEDBACK
        self._finished.set()

    @abstractmethod
    def _commit(self):
        """
        this method should be implemented by child classes to actually perform the commit
        """
        pass

    @abstractmethod
    def _rollback(self):
        """
        this method should be implemented by child classes to actually perform the rollback
        """
        pass


class InputTransactionScope(InputTransaction):
    """
    a helper class for reading several messages inside a transaction scope.
    """

    def __init__(self, device: 'InputDevice', with_transaction: bool = True):
        """
        :param device: the input device to read the messages from
        :param with_transaction: 'True' if we should actually use transactions, 'False' otherwise
        """
        super().__init__(device)
        self._with_transaction = with_transaction
        self._transactions: List[InputTransaction] = []

    def read_message(self,
                     cancellation_token: threading.Event,
                     timeout: Optional[float] = None) -> Optional['ReadResult']:
        """
        this method returns a message from the device, and adds its transaction to the transaction scope

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds (None means block until message is available).
         if the device doesn't have a message to return after timeout seconds, it will return None

        :return: a ReadResult object or None if no message was available.
        the device headers can contain extra information about the device that returned the message
        """

        read_result = self.device.read_message(cancellation_token=cancellation_token,
                                               timeout=timeout,
                                               with_transaction=self._with_transaction)

        if read_result is not None:
            self._transactions.append(read_result.transaction)

        return read_result

    def _commit(self):
        """
        commits all the transactions in scope
        """
        for transaction in self._transactions:
            if not transaction.finished:  # allows someone to rollback individual transactions within committed scope
                transaction.commit()
        self._transactions.clear()

    def _rollback(self):
        """
        rolls back all the transaction in scope
        """
        for transaction in self._transactions:
            if not transaction.finished:  # allows someone to commit individual transactions within rolled back scope
                transaction.rollback()
        self._transactions.clear()


class NULLTransaction(InputTransaction):
    """
    a transaction object that does nothing. used as placeholder for reading with with_transaction=False
    """

    def _commit(self):
        pass

    def _rollback(self):
        pass



import copy
from io import BytesIO
from typing import BinaryIO, Dict, Any, Optional, Union

MessageHeaders = Dict[str, Any]  # this is the type for message metadata


class Message:
    """
    this class is the basic unit that is read from, or sent to devices.
    """
    __slots__ = '_stream', '_headers'

    def __init__(self, data: Union[BinaryIO, bytes], headers: Optional[MessageHeaders] = None):
        """
        :param data: the bytes/stream containing the body of the message
        :param headers: (optional) headers containing metadata about the message
        """
        if isinstance(data, bytes):
            data = BytesIO(data)

        self._stream = data
        self._headers = headers or {}

    @property
    def stream(self) -> BinaryIO:
        """
        the stream for this message. notice that reading the stream, advances its position
        """
        return self._stream

    @property
    def bytes(self) -> bytes:
        """
        :return: the data of the message
        (reads the stream from current position to the end, than resets the position to the original value)
        """
        current_pos = self._stream.tell()
        data = self._stream.read()
        self._stream.seek(current_pos)
        return data

    @property
    def headers(self) -> MessageHeaders:
        """
        the headers for this message
        """
        return self._headers

    def copy(self, new_headers: Optional[MessageHeaders] = None):
        """
        makes a copy of the message, possibly giving it a new headers
        """
        stream_copy = copy.copy(self._stream)
        stream_copy.seek(0)
        if new_headers is None:
            new_headers = self._headers.copy()

        return Message(stream_copy, new_headers)

    def __eq__(self, other):
        if not isinstance(other, Message):
            return False

        return self.bytes == other.bytes and self._headers == other._headers

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo=None):
        stream_copy = copy.deepcopy(self._stream, memo)
        stream_copy.seek(0)
        return Message(stream_copy, copy.deepcopy(self._headers, memo))


DeviceHeaders = Dict[str, Any]  # this is the type for device specific headers, used to pass arguments to/from device


class MessageBundle:
    """
    this class holds a message and device headers (to get from device, or send to device)
    """
    __slots__ = "_message", "_device_headers"

    def __init__(self, message: Message, device_headers: Optional[DeviceHeaders] = None):
        """
        :param message: The Message.
        :param device_headers: Additional Headers that may return data from device, or affect its operation.
        """
        self._message = message
        self._device_headers = device_headers or {}

    @property
    def message(self) -> Message:
        """
        the message in this bundle
        """
        return self._message

    @property
    def device_headers(self) -> DeviceHeaders:
        """
        Additional Headers that may return data from device, or affect its operation.
        """
        return self._device_headers



import logging
import threading
from time import time
from typing import Optional

from messageflux.base_service import ServiceState
from messageflux.server_loop_service import ServerLoopService, LoopMetrics
from messageflux.utils import KwargsException


class LoopHealthAddonException(KwargsException):
    """
    an exception that is raised on LoopHealthAddon errors
    """
    pass


class LoopHealthAddon:
    """
    an addon that can be attached to a service, to make sure he is healthy.
    if the service gets stuck (service loop inactivity) or has to many consecutive failures,
    the addons stops the service
    """

    def __init__(self, *,
                 max_consecutive_failures: Optional[int] = None,
                 max_inactivity_timeout: Optional[float] = None):
        """

        :param max_consecutive_failures: the number of consecutive service loop failures,
        after which the service is stopped. None means to allow any number of failures
        :param max_inactivity_timeout: the number of seconds of inactivity (no loop ended events...) after which
        the service is stopped. None means to ignore inactivity.
        """

        self._max_consecutive_failures = max_consecutive_failures
        self._max_inactivity_timeout = max_inactivity_timeout
        self._cancellation_token = threading.Event()
        self._logger = logging.getLogger(__name__)
        self._service: Optional[ServerLoopService] = None

        self._consecutive_failures = 0
        self._last_loop_time = 0.0

    @property
    def service(self) -> Optional[ServerLoopService]:
        """
        the attached service (or None if no service is attached)
        """
        return self._service

    def attach(self, service: ServerLoopService) -> 'LoopHealthAddon':
        """
        attached the addon to a service, and returns the addon
        :param service: the service to attach to
        """
        if self._service is not None:
            raise LoopHealthAddonException('Cannot attach an already attached addon')

        self._consecutive_failures = 0
        self._last_loop_time = 0.0

        if self._max_inactivity_timeout is not None:
            if service.service_state in [ServiceState.STARTED, ServiceState.STARTING]:
                raise LoopHealthAddonException('Cannot monitor inactivity on an already started service')

            service.state_changed_event.subscribe(self._on_service_state_change)

        service.loop_ended_event.subscribe(self._on_loop_ended)
        self._service = service

        return self

    def detach(self):
        """
        detaches from a service
        """
        self._cancellation_token.set()
        service = self._service
        if service is None:
            return
        service.state_changed_event.unsubscribe(self._on_service_state_change)
        service.loop_ended_event.unsubscribe(self._on_loop_ended)
        self._service = None

    @property
    def consecutive_failures(self) -> int:
        """
        the number of consecutive failures the service has had
        """
        return self._consecutive_failures

    @property
    def last_loop_time(self) -> float:
        """
        the last time that the service loop has ended
        """
        return self._last_loop_time

    def _inactivity_watchdog_thread(self, service: ServerLoopService, max_inactivity_timeout: float):
        while not self._cancellation_token.is_set():
            timeout_time = self._last_loop_time + max_inactivity_timeout
            if time() >= timeout_time:
                self._logger.warning(
                    f'Service {service.name} exceeded {max_inactivity_timeout} '
                    f'seconds from last loop finished and will be stopped')
                service.stop()
                return

            time_to_sleep = timeout_time - time()
            self._cancellation_token.wait(time_to_sleep)

    def _on_service_state_change(self, service_state: ServiceState):
        if service_state == ServiceState.STARTED:
            self._on_service_started()
        elif service_state == ServiceState.STOPPING:
            self._on_service_stopping()

    def _on_service_started(self):
        if self._max_inactivity_timeout is not None:
            self._cancellation_token.clear()
            self._last_loop_time = time()
            t = threading.Thread(target=self._inactivity_watchdog_thread,
                                 args=[self._service, self._max_inactivity_timeout],
                                 daemon=True)
            t.start()

    def _on_service_stopping(self):
        self._cancellation_token.set()

    def _on_loop_ended(self, loop_metrics: LoopMetrics):
        self._last_loop_time = time()
        if loop_metrics.exception is None:
            self._consecutive_failures = 0
        else:
            self._consecutive_failures += 1

        if self._max_consecutive_failures is not None and self._consecutive_failures >= self._max_consecutive_failures:
            if self._service is not None:
                self._logger.warning(
                    f'Service {self._service.name} reached {self._consecutive_failures} and will be stopped')
                self._service.stop()



import logging
from abc import ABCMeta, abstractmethod
from typing import List, Tuple, Optional, Union, Iterable

from messageflux.iodevices.base import InputDevice, OutputDeviceManager, ReadResult, InputDeviceManager
from messageflux.iodevices.base.common import MessageBundle, Message
from messageflux.message_handling_service import MessageHandlingServiceBase

_logger = logging.getLogger(__name__)


class PipelineResult:
    """
    a result from pipeline handler
    """

    def __init__(self,
                 output_device_name: str,
                 message_bundle: Union[MessageBundle, Message]):
        self.output_device_name = output_device_name
        if isinstance(message_bundle, Message):
            message_bundle = MessageBundle(message=message_bundle)

        self.message_bundle = message_bundle


class PipelineHandlerBase(metaclass=ABCMeta):
    """
    a pipeline handler base class. used to handle a single message that passes through the pipeline
    """

    def prepare(self):
        """
        called when the service starts.
        can be overrided by child class to perform some initialization logic
        """
        pass

    def shutdown(self):
        """
        called when the service stops.
        can be overrided by child class to perform some cleanup logic
        """
        pass

    @abstractmethod
    def handle_message(self,
                       input_device: InputDevice,
                       message_bundle: MessageBundle) -> Optional[Union[PipelineResult, Iterable[PipelineResult]]]:
        """
        Handles a message from an input device. and returns a tuple of the output device name, message and headers to
        send to.

        :param input_device: The input device that sent the message.
        :param message_bundle: The message that was received.

        :return: None if the message should not be sent to any output device.
        PipelineResult (or iterable of) if messages should be sent to the output device with the given name.
        """
        pass


class FixedRouterPipelineHandler(PipelineHandlerBase):
    """
    A pipeline handler that routes all messages to a single output device. without any processing.
    """

    def __init__(self, output_device_name: str):
        self._output_device_name = output_device_name

    def handle_message(self,
                       input_device: InputDevice,
                       message_bundle: MessageBundle) -> Optional[PipelineResult]:
        """
        routes the message to the fixed output

        :param input_device: The input device that sent the message.
        :param message_bundle: The message that was received.

        :return: a pipeline result that has the input message bundle, and the fixed output device name
        """
        output_message_bundle = MessageBundle(message=message_bundle.message.copy(), device_headers={})
        return PipelineResult(output_device_name=self._output_device_name,
                              message_bundle=output_message_bundle)


class PipelineService(MessageHandlingServiceBase):
    """
    a service that uses a PipelineHandler object to process messages from input, and send to output
    """

    def __init__(self, *,
                 input_device_manager: InputDeviceManager,
                 input_device_names: Union[List[str], str],
                 pipeline_handler: PipelineHandlerBase,
                 output_device_manager: Optional[OutputDeviceManager] = None,
                 **kwargs):
        """

        :param pipeline_handler: the pipeline handler to use for handling the messages
        :param output_device_manager: Optional. the output device manager to send messages to
        :param **kwargs: passed to parent as is
        """
        super().__init__(input_device_manager=input_device_manager,
                         input_device_names=input_device_names,
                         **kwargs)
        self._output_device_manager = output_device_manager
        self._pipeline_handler = pipeline_handler

    def _prepare_service(self):
        super()._prepare_service()
        if self._output_device_manager is not None:
            self._output_device_manager.connect()
        self._pipeline_handler.prepare()

    def _handle_message_batch(self, batch: List[Tuple[InputDevice, ReadResult]]):
        for input_device, read_result in batch:
            pipeline_handler_result = self._pipeline_handler.handle_message(input_device, read_result)
            if pipeline_handler_result is not None:
                if isinstance(pipeline_handler_result, PipelineResult):
                    pipeline_handler_result = [pipeline_handler_result]
                for pipeline_result in pipeline_handler_result:
                    if self._output_device_manager is None:
                        _logger.warning("pipeline handler returned a result to output to device: "
                                        f"'{pipeline_result.output_device_name}', "
                                        f"but no output_device_manager was given")
                        continue

                    output_device = self._output_device_manager.get_output_device(pipeline_result.output_device_name)
                    output_device.send_message(message=pipeline_result.message_bundle.message,
                                               device_headers=pipeline_result.message_bundle.device_headers)

    def _finalize_service(self, exception: Optional[Exception] = None):
        try:
            super()._finalize_service(exception)
        finally:
            if self._output_device_manager is not None:
                self._output_device_manager.disconnect()
            self._pipeline_handler.shutdown()



import logging
import threading
import time
from multiprocessing.process import BaseProcess
from typing import List, Optional

from messageflux.base_service import BaseService
from messageflux.multiprocessing.singleprocesshandler import ServiceFactory, SingleProcessHandler


class MultiProcessRunner(BaseService):
    """
    a class that runs multiple services in processes. it handles the child process, and check liveness
    """

    def __init__(self, *,
                 service_factory: ServiceFactory,
                 instance_count: int,
                 shutdown_timeout: int = 5,
                 live_check_interval: int = 60,
                 live_check_timeout: int = 10,
                 restart_on_failure: bool = True,
                 **kwargs):
        """
            :param service_factory: a factory class that creates the service to run in a different process
            :param instance_count: the number of processes to run
            :param shutdown_timeout: the time (seconds) to wait after calling 'stop' to violently stop the subprocesses
            :param live_check_interval: the interval in seconds to send the liveness message to queue
            :param live_check_timeout: the time to wait for the process answer to liveness test
            :param restart_on_failure: should we restart a process if it fails?
        """

        super().__init__(**kwargs)
        self._service_factory = service_factory
        self._instance_count = instance_count
        self._shutdown_timeout = shutdown_timeout
        self._live_check_interval = live_check_interval
        self._live_check_timeout = live_check_timeout
        self._restart_on_failure = restart_on_failure
        self._logger = logging.getLogger(__name__)

        self._process_handlers: List[SingleProcessHandler] = []

    def _is_alive(self) -> bool:
        for handler in self._process_handlers:
            if handler.is_alive():
                return True

        return False

    def _run_service(self, cancellation_token: threading.Event):
        for i in range(self._instance_count):
            self._logger.info(f'Starting service instance {i}')
            self._run_service_instance(i)

    @property
    def processes(self) -> List[BaseProcess]:
        """
        the list of child processes that runs the service instances
        """
        return [handler.process for handler in self._process_handlers if handler.process is not None]

    def _run_service_instance(self, instance_index: int):
        handler = SingleProcessHandler(service_factory=self._service_factory,
                                       instance_index=instance_index,
                                       total_instances=self._instance_count,
                                       live_check_interval=self._live_check_interval,
                                       live_check_timeout=self._live_check_timeout)
        handler.start(self._on_handler_exit)
        self._process_handlers.append(handler)

    def _on_handler_exit(self, handler: SingleProcessHandler):
        if not self._restart_on_failure:
            return

        if self._cancellation_token.is_set():
            return  # no need to restart if we are shutting down

        if handler in self._process_handlers:
            self._process_handlers.remove(handler)

        self._run_service_instance(handler.instance_index)

    def _finalize_service(self, exception: Optional[Exception] = None):
        super()._finalize_service(exception)
        for handler in self._process_handlers:
            handler.stop()
        time.sleep(self._shutdown_timeout)
        still_running = [handler.process for handler in self._process_handlers if
                         handler.is_alive() and handler.process is not None]
        if still_running:
            self._logger.warning(f'{len(still_running)} processes still running after {self._shutdown_timeout} seconds')
            for handler_process in still_running:
                handler_process.terminate()
            time.sleep(self._shutdown_timeout)
            still_running = [handler.process for handler in self._process_handlers if
                             handler.is_alive() and handler.process is not None]
            if still_running:
                self._logger.warning(
                    f'{len(still_running)} processes still running after {self._shutdown_timeout} seconds')
                for handler_process in still_running:
                    handler_process.kill()
                time.sleep(self._shutdown_timeout)
                still_running = [handler.process for handler in self._process_handlers if
                                 handler.is_alive() and handler.process is not None]
                if still_running:
                    self._logger.error(
                        f'{len(still_running)} processes still running after {self._shutdown_timeout} seconds')
                    pass


def get_service_runner(*,
                       service_factory: ServiceFactory,
                       instance_count: int,
                       shutdown_timeout: int = 5,
                       live_check_interval: int = 60,
                       live_check_timeout: int = 10,
                       restart_on_failure: bool = True,
                       **kwargs) -> BaseService:
    """
    a helper method, that the creates the MultiProcessRunner if instance_count is greater then 1.
    otherwise, it just returns the service itself.

    :param service_factory: a factory class that creates the service to run in a different process
    :param instance_count: the number of processes to run
    :param shutdown_timeout: the time (seconds) to wait after calling 'stop' to violently stop the subprocesses
    :param live_check_interval: the interval in seconds to send the liveness message to queue
    :param live_check_timeout: the time to wait for the process answer to liveness test
    :param restart_on_failure: should we restart a process if it fails?
    """
    if instance_count <= 1:
        return service_factory.create_service(instance_index=0, total_instances=1)
    else:
        return MultiProcessRunner(service_factory=service_factory,
                                  instance_count=instance_count,
                                  shutdown_timeout=shutdown_timeout,
                                  live_check_interval=live_check_interval,
                                  live_check_timeout=live_check_timeout,
                                  restart_on_failure=restart_on_failure,
                                  **kwargs)



import os
from typing import Optional, Dict, Any

from messageflux.iodevices.base import OutputDeviceManager, Message
from messageflux.logging.bulk_rotating_handler_base import BulkRotatingHandlerBase


class BulkRotatingDeviceHandler(BulkRotatingHandlerBase):
    """
    Handler for logging into an output device when the current file reaches a certain size or time.
    """

    def __init__(self,
                 live_log_path: str,
                 output_device_manager: OutputDeviceManager,
                 output_device_name: str,
                 metadata: Optional[Dict[str, Any]] = None,
                 bkp_log_path: Optional[str] = None,
                 max_records: int = 1000,
                 max_time: int = 60,
                 live_log_prefix: str = ''):
        """
        Open a log file and use it as the stream for logging.
        when max_records are written, or max_time has passed, a rollover occurs
        rollover copies the live log from live_log_path to rotated_log_path

        :param live_log_path: the path to write the live log to
        :param output_device_manager: the device manager to send the logs to
        :param output_device_name: the name of the output_device to send to
        :param metadata: a dictionary of metadata to send on the device along with the log
        :param bkp_log_path: the path to write to rotated log to, if writing to rotated_log_path fails
        :param max_records: the maximum number of records to write before rotation
        :param max_time: the maximum time (in seconds) to wait before rotation
        :param live_log_prefix: the prefix for live log file
        """
        self._output_device_manager = output_device_manager
        self._output_device_name = output_device_name

        self._output_device_manager.connect()
        self._output_device = self._output_device_manager.get_output_device(name=self._output_device_name)
        self._metadata = metadata or {}

        super(BulkRotatingDeviceHandler, self).__init__(live_log_path=live_log_path,
                                                        bkp_log_path=bkp_log_path,
                                                        max_records=max_records,
                                                        max_time=max_time,
                                                        live_log_prefix=live_log_prefix)

    def _move_log_to_destination(self, src_file: str):
        """
        this moves the live log from a file, to its destination (the rotated log path)
        """
        with open(src_file, 'rb') as log_file:
            self._output_device.send_message(Message(log_file, self._metadata.copy()))

        os.remove(src_file)

    def close(self):
        """
        closes the handler (disconnects from the output device)
        """
        try:
            super(BulkRotatingDeviceHandler, self).close()
        finally:
            self._output_device_manager.disconnect()



import threading
from abc import ABCMeta, abstractmethod
from typing import Optional

from messageflux.iodevices.base import InputDevice, ReadResult, InputDeviceManager


class InputTransformerBase(metaclass=ABCMeta):
    """
    transformer for input devices
    """

    def connect(self):
        """
        this method is called when the transformer device manager is connected
        """
        pass

    def disconnect(self):
        """
        this method is called when the transformer device manager is disconnected
        """
        pass

    @abstractmethod
    def transform_incoming_message(self,
                                   input_device: 'TransformerInputDevice',
                                   read_result: ReadResult) -> ReadResult:
        """
        Transform the message that was received from the underlying device.

        :param input_device: the input device that the transformer runs on
        :param read_result: the original ReadResult received from the underlying device
        :return: the transformed ReadResult
        """
        pass


class TransformerInputDevice(InputDevice['TransformerInputDeviceManager']):
    """
    an input device that transforms the input
    """

    def __init__(self,
                 manager: 'TransformerInputDeviceManager',
                 name: str,
                 inner_device: InputDevice,
                 transformer: InputTransformerBase):
        """

        :param manager: the input device manager
        :param name: the name of this device
        :param inner_device: the inner device that it wraps
        :param transformer: the transformer to use to transform the incoming messages
        """
        super(TransformerInputDevice, self).__init__(manager, name)
        self._transformer = transformer
        self._inner_device = inner_device

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional[ReadResult]:
        read_result = self._inner_device.read_message(cancellation_token=cancellation_token,
                                                      timeout=timeout,
                                                      with_transaction=with_transaction)
        if read_result is not None:
            read_result = self._transformer.transform_incoming_message(self, read_result)

        return read_result

    def close(self):
        """
        closes the inner device
        """
        self._inner_device.close()


class TransformerInputDeviceManager(InputDeviceManager[TransformerInputDevice]):
    """
    a wrapper input device manager, that wraps the devices in transformer input devices
    """

    def __init__(self, inner_device_manager: InputDeviceManager, transformer: InputTransformerBase):
        """

        :param inner_device_manager: the inner device manager
        :param transformer: the input transformer to use
        """
        self._inner_device_manager = inner_device_manager
        self._transformer = transformer

    def connect(self):
        """
        connects the inner device manager and the transformer
        """
        self._transformer.connect()
        self._inner_device_manager.connect()

    def disconnect(self):
        """
        disconnects the inner device manager and the transformer
        """
        self._inner_device_manager.disconnect()
        self._transformer.disconnect()

    def get_input_device(self, name: str) -> TransformerInputDevice:
        """
        returns a wrapped input device

        :param name: the name of the input device to get
        """

        inner_input_device = self._inner_device_manager.get_input_device(name)
        return TransformerInputDevice(self, name, inner_input_device, self._transformer)



import threading
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
# noinspection PyPackageRequirements
from contextvars import ContextVar
from typing import Dict, Any

_context_var: ContextVar = ContextVar('__asyncio_context__', default={})


class Context(metaclass=ABCMeta):
    """
    a stack like context object. allows to store some key-value data and access it
    """

    @abstractmethod
    def get(self, key: str, default=None) -> Any:
        """
        gets the current value for 'key' in the context. returns default if key does not exist

        :param key: the key to get
        :param default: the default value to return if key does not exist
        :return: the current value for 'key' in the context. returns 'default' if key does not exist
        """
        pass

    @abstractmethod
    @contextmanager
    def start_context(self, **kwargs):
        """
        starts the context manager

        :param **kwargs: kwargs to push to the context
        """
        pass

    @abstractmethod
    def get_current_context(self) -> Dict[str, Any]:
        """
        gets the complete current context
        """
        pass


class _GlobalContext(Context):
    """
    this class is a global context. same for everyone
    """

    def __init__(self):
        self._state = {}

    def _push(self, key: str, value: Any):
        stack = self._state.setdefault(key, [])
        stack.append(value)

    def _pop(self, key: str) -> Any:
        stack = self._state.get(key)
        assert stack is not None
        value = stack.pop()
        if not stack:
            self._state.pop(key, None)
        return value

    def get(self, key: str, default=None) -> Any:
        """
        gets the current value for 'key' in the context. returns default if key does not exist

        :param key: the key to get
        :param default: the default value to return if key does not exist
        :return: the current value for 'key' in the context. returns 'default' if key does not exist
        """
        stack = self._state.get(key, [default])
        return stack[-1]

    def get_current_context(self) -> Dict[str, Any]:
        """
        gets the complete current context
        """
        return {key: stack[-1] for key, stack in self._state.items()}

    @contextmanager
    def start_context(self, **kwargs):
        """
        starts the context manager

        :param kwargs: kwargs to push to the context
        """
        for key, value in kwargs.items():
            self._push(key, value)
        try:
            yield self
        finally:
            for key in kwargs.keys():
                self._pop(key)


class _ThreadLocalContext(threading.local, _GlobalContext):
    """
    this class is a thread local context. each thread sees a different context stack
    """

    pass


class _AsyncioContext(Context):
    """
    this context is for asyncio. each asyncio chain sees a different context stack
    """

    def get(self, key: str, default=None) -> Any:
        """
        gets the current value for 'key' in the context. returns default if key does not exist

        :param key: the key to get
        :param default: the default value to return if key does not exist
        :return: the current value for 'key' in the context. returns 'default' if key does not exist
        """
        return _context_var.get().get(key, default)

    def get_current_context(self) -> Dict[str, Any]:
        """
        gets the complete current context
        """
        return _context_var.get().copy()

    @contextmanager
    def start_context(self, **kwargs):
        """
        starts the context manager

        :param kwargs: kwargs to push to the context
        """

        new_context = _context_var.get().copy()
        new_context.update(kwargs)
        token = _context_var.set(new_context)
        try:
            yield self
        finally:
            _context_var.reset(token)


global_context: Context = _GlobalContext()
thread_local_context: Context = _ThreadLocalContext()
asyncio_context: Context = _AsyncioContext()



import json
import logging
import os
import threading
from hashlib import md5
from threading import Event
from typing import Optional, Dict, Any

from time import time

from messageflux.iodevices.rabbitmq.rabbitmq_poison_counting_input_device import PoisonCounterBase


class FileSystemPoisonCounter(PoisonCounterBase):
    """
    a poison counter that uses a shared folder to coordinate the consumers

    it is a reasonable assumption, that at any given moment, a message is only handled by one consumer,
    since the queue takes care of that.
    therefore, there should be no race on reading/writing the file (not including cases of hash collision...)
    """
    COUNTER_FILE_SUFFIX = '.PMC'
    MESSAGE_ID_PROP_NAME = 'message_id'
    COUNTER_PROP_NAME = 'counter'

    def __init__(self, base_folder: str, file_cleanup_timeout: int = 600):
        """
        :param base_folder: the shared folder path, to use for counter files.
        it is recommended that it will be unique per consumer group
        :param file_cleanup_timeout: the time (in seconds), after which, the counter file is considered old,
        and should be deleted
        """
        self._base_folder = base_folder
        self._file_cleanup_timeout = file_cleanup_timeout

        os.makedirs(self._base_folder, exist_ok=True)

        self._should_stop = Event()
        self._cleanup_thread: Optional[threading.Thread] = None

    def start(self):
        """
        starts the cleanup thread
        """
        self._should_stop.clear()
        if self._cleanup_thread is None:
            self._cleanup_thread = threading.Thread(target=self._do_cleanup_thread, daemon=True)
            self._cleanup_thread.start()

    def stop(self):
        """
        stops the cleanup thread
        """
        self._should_stop.set()
        if self._cleanup_thread is not None:
            self._cleanup_thread.join()
            self._cleanup_thread = None

    def _do_cleanup_thread(self):
        logger = logging.getLogger(__name__)
        while not self._should_stop.is_set():
            try:
                now = time()
                for direntry in os.scandir(self._base_folder):
                    try:
                        if not direntry.is_file():
                            continue
                        counter_file = direntry.path
                        if not counter_file.endswith(self.COUNTER_FILE_SUFFIX):
                            continue
                        file_age = now - direntry.stat().st_mtime
                        if file_age < self._file_cleanup_timeout:
                            # don't remove counter files that are too new (might be a file of another pod/process)
                            continue

                        os.remove(counter_file)
                    except FileNotFoundError:
                        continue  # Couldn't handle the file, probably because someone else got to it first
            except Exception:
                logger.warning('Error in cleanup thread', exc_info=True)
            # sleep for at least a few seconds so we won't spam
            self._should_stop.wait(self._file_cleanup_timeout)

    def _get_counter_filepath(self, message_id: str):
        """
        calculates the full path of the counter file
        """
        filename = md5(message_id.encode('utf8')).hexdigest()
        return os.path.join(self._base_folder, f'{filename}{self.COUNTER_FILE_SUFFIX}')

    def increment_and_return_counter(self, message_id: str) -> int:
        """
        reads the counter file, increments the counter, and write it back to counter file
        """
        counter_filepath = self._get_counter_filepath(message_id)
        data: Dict[str, Any] = {self.MESSAGE_ID_PROP_NAME: message_id,
                                self.COUNTER_PROP_NAME: 0}

        try:
            if os.path.exists(counter_filepath):
                with open(counter_filepath, 'r') as f:
                    data = json.load(f)
        except FileNotFoundError:  # someone deleted the file as we read it... NOT SUPPOSED TO HAPPEN
            pass

        data[self.MESSAGE_ID_PROP_NAME] = message_id  # THIS IS ONLY FOR SAFETY... PROBABLY REDUNDANT
        data[self.COUNTER_PROP_NAME] += 1

        with open(counter_filepath, 'w') as f:
            json.dump(data, f)
        os.chmod(counter_filepath, 0o777)

        return data[self.COUNTER_PROP_NAME]

    def delete_counter(self, message_id: str):
        """
        deletes the counter file
        """
        try:
            counter_filepath = self._get_counter_filepath(message_id)
            if os.path.exists(counter_filepath):
                os.remove(counter_filepath)
        except FileNotFoundError:  # file already deleted... so we don't care
            pass



import logging
import multiprocessing
import os
import threading
from abc import ABCMeta, abstractmethod
from multiprocessing import process
from multiprocessing.process import BaseProcess
from typing import Optional, Callable, TYPE_CHECKING

from messageflux.base_service import BaseService

if TYPE_CHECKING:
    from multiprocessing.connection import _ConnectionBase


class ServiceFactory(metaclass=ABCMeta):
    """
    This class is used to create a service instance.
    """

    @abstractmethod
    def create_service(self, instance_index: int, total_instances: int) -> BaseService:
        """
        creates the service instance. this will run in the child service

        :param instance_index: the instance index (0 based) out of all the indexes
        :param total_instances: the total number of indexes
        :return: an instance of BaseService
        """
        pass


_STOP_MESSAGE = 'STOP'
_TEST_ALIVE_MESSAGE = 'TEST_ALIVE'

INSTANCE_INDEX_ENV_VAR = 'MULTI_PROCESS_INSTANCE_INDEX'
INSTANCE_COUNT_ENV_VAR = 'MULTI_PROCESS_INSTANCE_COUNT'

_logger = logging.getLogger(__name__)


def _start_service_and_listen_queue(service_factory: ServiceFactory,
                                    child_pipe: '_ConnectionBase',
                                    instance_index: int,
                                    total_instances: int):
    try:
        os.environ[INSTANCE_INDEX_ENV_VAR] = str(instance_index)
        os.environ[INSTANCE_COUNT_ENV_VAR] = str(total_instances)

        service = service_factory.create_service(instance_index=instance_index, total_instances=total_instances)

        def _listen_to_pipe(inner_service: BaseService, pipe: '_ConnectionBase'):
            try:
                while True:
                    message = pipe.recv()
                    if message == _TEST_ALIVE_MESSAGE:
                        pipe.send(inner_service.is_alive)
                    elif message == _STOP_MESSAGE:
                        _logger.info(f'Stopping Service #{instance_index}')
                        break
                    else:
                        _logger.error(f'Unknown message received: {message}')
                        pass
            except Exception:
                _logger.exception('Error in pipe listener')
                pass
            finally:
                inner_service.stop()
                pipe.close()

        t = threading.Thread(target=_listen_to_pipe, args=(service, child_pipe), daemon=True)
        t.start()
        _logger.info(f'Starting Service #{instance_index}')
        service.start()
    except Exception:
        _logger.exception(f'Error starting service #{instance_index}')
        raise


class SingleProcessHandler:
    """
    This class is used to handle a single process.
    """

    def __init__(self,
                 service_factory: ServiceFactory,
                 instance_index: int,
                 total_instances: int,
                 live_check_interval: int = 60,
                 live_check_timeout: int = 10):
        self._service_factory = service_factory
        self._live_check_interval = live_check_interval
        self._live_check_timeout = live_check_timeout
        self._liveness_thread: Optional[threading.Thread] = None
        self._stop_was_called = threading.Event()
        self._parent_pipe: Optional['_ConnectionBase'] = None
        self._process: Optional[process.BaseProcess] = None
        self._instance_index = instance_index
        self._total_instances = total_instances

    @property
    def instance_index(self) -> int:
        """
        the index of this service instance
        """
        return self._instance_index

    def _liveness_check(self):
        while not self._stop_was_called.wait(self._live_check_interval):
            alive = False
            if self._parent_pipe is not None and not self._parent_pipe.closed:
                self._parent_pipe.send(_TEST_ALIVE_MESSAGE)
                has_answer = self._parent_pipe.poll(self._live_check_timeout)
                if has_answer:
                    alive = self._parent_pipe.recv()

            if not alive:
                _logger.error(f'Service #{self._instance_index} is not alive. Stopping.')
                self.stop()
                self.kill()
                break

    def start(self, exit_callback: Callable[['SingleProcessHandler'], None]) -> threading.Thread:
        """
        Start the service.

        :param exit_callback: a callback to call when the process has exited
        :return: a thread that runs that instance.
        """

        context = multiprocessing.get_context('spawn')
        child_pipe: '_ConnectionBase'
        self._parent_pipe, child_pipe = context.Pipe()
        self._stop_was_called.clear()

        self._process = context.Process(target=_start_service_and_listen_queue,
                                        args=(
                                            self._service_factory, child_pipe, self._instance_index,
                                            self._total_instances))

        def _run_process():
            assert self._process is not None
            self._process.start()
            self._process.join()
            self._cleanup()

            exit_callback(self)

        run_process_thread = threading.Thread(target=_run_process, daemon=True)
        run_process_thread.start()

        self._liveness_thread = threading.Thread(target=self._liveness_check, daemon=True)
        self._liveness_thread.start()

        return run_process_thread

    def stop(self):
        """
        requests the child process to stop
        """
        self._stop_was_called.set()
        if self._parent_pipe is not None:
            self._parent_pipe.send(_STOP_MESSAGE)

    def is_alive(self) -> bool:
        """
        returns whether the child process is alive
        :return: True if the child process is alive, False otherwise
        """
        return self._process is not None and self._process.is_alive()

    @property
    def pid(self) -> Optional[int]:
        """
        the pid of the child process (or None if no child process exists)
        """
        if self._process is None:
            return None
        return self._process.pid

    @property
    def process(self) -> Optional[BaseProcess]:
        """
        the child process object for this instance (if exists.)
        """
        return self._process

    def terminate(self):
        """
        tries to forcefully terminate the child process
        """
        if self.is_alive():
            assert self._process is not None
            self._process.terminate()

    def kill(self):
        """
        tries to forcefully KILL the child process
        """
        if self.is_alive():
            assert self._process is not None
            self._process.kill()

    def _cleanup(self):
        if self._parent_pipe is not None:
            self._parent_pipe.close()
            self._parent_pipe = None



import logging
import os
import shutil
from typing import Optional

from messageflux.iodevices.base import OutputDeviceManager, OutputDevice, OutputDeviceException
from messageflux.iodevices.base.common import MessageBundle, Message
from messageflux.iodevices.file_system.file_system_device_manager_base import FileSystemDeviceManagerBase
from messageflux.iodevices.file_system.file_system_serializer import FileSystemSerializerBase, \
    DefaultFileSystemSerializer
from messageflux.metadata_headers import MetadataHeaders
from messageflux.utils import get_random_id


class FileSystemOutputDevice(OutputDevice['FileSystemOutputDeviceManager']):
    """
    And OutputDevice that writes to filesystem
    """

    def __init__(self,
                 manager: 'FileSystemOutputDeviceManager',
                 name: str,
                 tmp_folder: str,
                 queues_folder: str,
                 format_filename: Optional[str] = None,
                 serializer: Optional[FileSystemSerializerBase] = None):
        """
        :param manager: the output device manager that created this device
        :param name: the name of this device
        :param tmp_folder: the folder to create temporary files in
        :param queues_folder: the base folder for all queues
        :param format_filename: the filename format to save the product, "{filename}-{item_id}"
        :param serializer: the serializer to use
        """
        super(FileSystemOutputDevice, self).__init__(manager, name)
        self._tmp_folder = tmp_folder
        self._output_folder = os.path.join(queues_folder, name)
        try:
            os.makedirs(self._tmp_folder, exist_ok=True)
            os.makedirs(self._output_folder, exist_ok=True)
        except Exception as e:
            raise OutputDeviceException('Error creating output device') from e

        self._format = format_filename
        self._logger = logging.getLogger(__name__)
        self._serializer = serializer or DefaultFileSystemSerializer()

    def _send_message(self, message_bundle: MessageBundle):
        """
        sends a message to the device

        :param message_bundle: the message bundle to send
        """
        item_id = message_bundle.device_headers.get(MetadataHeaders.ITEM_ID, '')
        self._logger.debug('Sending product to {} device'.format(self.name))
        self._save_to_folder(item_id, message_bundle.message)

    def _save_to_folder(self, item_id: str, message: Message):
        """
        saves the given message to the appropriate folder

        :param message: the message tp be saved
        """
        try:
            filename = ''
            try:
                if self._format:
                    filename = self._format.format(**message.headers)
            except KeyError:
                self._logger.error(
                    'Filename format isn\'t satisfied, format:{}, headers:{}, puts default filename instead'.format(
                        self._format, message.headers))
            if not filename:
                if item_id:
                    item_id += '-'
                filename = "{itemid}{uuid}.SBM".format(itemid=item_id if item_id else str(item_id),
                                                       uuid=get_random_id())
            tmp_fullpath = os.path.join(self._tmp_folder, filename)
            with open(tmp_fullpath, 'wb') as f:
                stream_to_write = self._serializer.serialize(message=message)
                f.write(stream_to_write.read())

            os.chmod(tmp_fullpath, 0o777)
            final_fullpath = os.path.join(self._output_folder, filename)
            shutil.move(tmp_fullpath, final_fullpath)
            self._logger.debug(f'Wrote product to path {final_fullpath}')
        except Exception as e:
            raise OutputDeviceException('Error writing to device') from e


class FileSystemOutputDeviceManager(FileSystemDeviceManagerBase, OutputDeviceManager[FileSystemOutputDevice]):
    """
    this is a device manager that generates output filesystem devices
    """

    def __init__(self,
                 root_folder: str,
                 queue_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_QUEUES_SUB_DIR,
                 tmp_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_TMPDIR_SUB_DIR,
                 bookkeeping_dir_name: str = FileSystemDeviceManagerBase.DEFAULT_BOOKKEEPING_SUB_DIR,
                 serializer: Optional[FileSystemSerializerBase] = None,
                 output_filename_format: Optional[str] = None):
        """
        :param root_folder: the root folder to use for the manager
        :param queue_dir_name: the name of the subdirectory under root_folder that holds the queues
        :param tmp_dir_name: the name of the subdirectory under root_folder to use for temp files
        :param bookkeeping_dir_name: the name of the subdirectory under root_folder that holds the book-keeping data
        :param serializer: the serializer to use to write messages to files. None will use the default serializer
        :param output_filename_format: the filename format to save a product, "{filename}-{item_id}"
        """
        super(FileSystemOutputDeviceManager, self).__init__(root_folder=root_folder,
                                                            queue_dir_name=queue_dir_name,
                                                            tmp_dir_name=tmp_dir_name,
                                                            bookkeeping_dir_name=bookkeeping_dir_name,
                                                            serializer=serializer)
        self._output_filename_format = output_filename_format

    def connect(self):
        """
        connects to the device manager
        """
        try:
            self._create_all_directories()
        except Exception as ex:
            raise OutputDeviceException('Error connection to Device Manager') from ex

    def get_output_device(self, name: str) -> FileSystemOutputDevice:
        """
        Returns an outgoing device by name

        :param name: the name of the device to write to
        :return: an outgoing device for 'name'
        """
        try:
            return FileSystemOutputDevice(manager=self,
                                          name=name,
                                          tmp_folder=self._tmp_folder,
                                          queues_folder=self._queues_folder,
                                          format_filename=self._output_filename_format,
                                          serializer=self._serializer)
        except Exception as e:
            raise OutputDeviceException("Error getting output device") from e



import logging
import os
import socket
import ssl
import threading
from io import BytesIO
from typing import TYPE_CHECKING, Dict, Optional, Tuple, Union, List, Any

from messageflux.iodevices.base import InputDevice, InputTransaction, ReadResult, InputDeviceException, Message, \
    InputDeviceManager
from messageflux.iodevices.base.input_transaction import NULLTransaction
from messageflux.iodevices.rabbitmq.rabbitmq_device_manager_base import RabbitMQDeviceManagerBase
from messageflux.utils import ThreadLocalMember

try:
    from pika import spec
except ImportError as ex:
    raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from ex

if TYPE_CHECKING:
    from pika.adapters.blocking_connection import BlockingChannel


class RabbitMQInputTransaction(InputTransaction):
    """
    represents a InputTransaction for RabbitMQ
    """

    def __init__(self,
                 cancellation_token: threading.Event,
                 device: 'RabbitMQInputDevice',
                 channel: 'BlockingChannel',
                 delivery_tag: int):
        """

        :param cancellation_token: the cancellation token, to check before acking
        :param device: the device that returned this transaction
        :param channel: the BlockingChannel that the item was read from
        :param delivery_tag: the delivery tag for this item
        """
        super(RabbitMQInputTransaction, self).__init__(device=device)
        self._cancellation_token = cancellation_token
        self._channel = channel
        self._delivery_tag = delivery_tag
        self._logger = logging.getLogger(__name__)

    @property
    def channel(self) -> 'BlockingChannel':
        """
        the channel that the item was read from
        """
        return self._channel

    @property
    def delivery_tag(self) -> int:
        """
        the delivery tag for this item
        """
        return self._delivery_tag

    def _commit(self):
        try:
            if self._cancellation_token.is_set():
                try:
                    self._channel.cancel()  # cancel the consumer before acking
                except Exception:
                    pass
            self._channel.basic_ack(self._delivery_tag)
        except Exception:
            self._logger.warning('commit failed', exc_info=True)

    def _rollback(self):
        try:
            if self._cancellation_token.is_set():
                try:
                    self._channel.cancel()  # cancel the consumer before nacking
                except Exception:
                    pass
            self._channel.basic_nack(self._delivery_tag, requeue=True)
        except Exception:
            self._logger.warning('rollback failed', exc_info=True)


class RabbitMQInputDevice(InputDevice['RabbitMQInputDeviceManager']):
    """
    represents an RabbitMQ input device
    """

    _channel: Union[ThreadLocalMember[Optional['BlockingChannel']], Optional['BlockingChannel']] = ThreadLocalMember(
        init_value=None)

    @staticmethod
    def _get_rabbit_headers(method_frame, header_frame):
        return {
            "exchange": method_frame.exchange,
            "routing_key": method_frame.routing_key,
            "content_type": header_frame.content_type,
            "content_encoding": header_frame.content_encoding,
            "priority": header_frame.priority,
            "correlation_id": header_frame.correlation_id,
            "reply_to": header_frame.reply_to,
            "expiration": header_frame.expiration,
            "message_id": header_frame.message_id,
            "timestamp": header_frame.timestamp,
            "type": header_frame.type,
            "user_id": header_frame.user_id,
            "app_id": header_frame.app_id
        }

    def __init__(self,
                 device_manager: 'RabbitMQInputDeviceManager',
                 queue_name: str,
                 consumer_args: Optional[Dict[str, str]] = None,
                 prefetch_count: int = 1,
                 use_consumer: bool = True):
        """
        constructs a new input RabbitMQ device

        :param device_manager: the RabbitMQ device Manager that holds this device
        :param queue_name: the name for the queue
        :param consumer_args: the arguments to create the consumer with
        only relevent if "use_consumer" is True
        :param int prefetch_count: the number of unacked messages that can be consumed
        only relevent if "use_consumer" is True
        :param bool use_consumer: True to use the 'consume' method, False to use 'basic_get'
        """
        super().__init__(device_manager, queue_name)
        self._device_manager = device_manager
        self._queue_name = queue_name
        self._logger = logging.getLogger(__name__)
        if consumer_args is None:
            consumer_args = {'hostname': socket.gethostname(), 'PID': str(os.getpid())}

        self._consumer_args = consumer_args
        self._prefetch_count = max(1, prefetch_count)
        self._use_consumer = use_consumer
        self._last_consumer_auto_ack: Optional[bool] = None

    def _reconnect_device_manager(self):
        """
        reconnects the RabbitMQ device manager
        """
        try:
            if self._channel is not None and self._channel.is_open:
                assert self._channel is not None
                if self._use_consumer:
                    try:
                        self._channel.cancel()
                    except Exception:
                        pass
                self._channel.close()
            self._channel = self._device_manager.connection.channel()

            assert self._channel is not None
            self._channel.basic_qos(prefetch_count=self._prefetch_count)
        except Exception as e:
            raise InputDeviceException('Could not connect to rabbitmq.') from e

    def _get_channel(self) -> 'BlockingChannel':
        """
        gets a channel
        """
        if self._channel is None or not self._channel.is_open:
            self._reconnect_device_manager()

        assert self._channel is not None
        return self._channel

    def _get_data_from_queue(self,
                             cancellation_token: threading.Event,
                             timeout: Optional[float],
                             with_transaction: bool) -> Optional['ReadResult']:
        """
        performs a single read from queue

        :param cancellation_token: the cancellation token, to pass to transaction
        :param timeout: the timeout in seconds to block. negative number means no blocking
        :param with_transaction: does this read is to be done with transaction?
        :return: the stream and metadata, or None,None if no message in queue
        """
        channel = self._get_channel()
        get_timeout: Optional[float] = None
        if timeout is not None:
            get_timeout = max(0.01, timeout)
        body: Optional[bytes]
        header_frame: Optional[spec.BasicProperties]
        method_frame: Optional[Union[spec.Basic.Deliver, spec.Basic.GetOk]]

        body, header_frame, method_frame = self._get_frames_from_queue(channel,
                                                                       get_timeout,
                                                                       with_transaction=with_transaction)
        if method_frame is None:  # no message in queue
            return None

        assert body is not None
        assert header_frame is not None

        return self._create_response_from_frames(cancellation_token=cancellation_token,
                                                 body=body,
                                                 header_frame=header_frame,
                                                 method_frame=method_frame,
                                                 channel=channel,
                                                 with_transaction=with_transaction)

    def _create_response_from_frames(self,
                                     cancellation_token: threading.Event,
                                     body: bytes,
                                     header_frame: spec.BasicProperties,
                                     method_frame: Union[spec.Basic.Deliver, spec.Basic.GetOk],
                                     channel: 'BlockingChannel',
                                     with_transaction: bool) -> ReadResult:
        """
        creates the read result from the data returned from rabbitmq

        :param cancellation_token: the cancellation token, to pass to transaction
        :param body: the body of the message
        :param header_frame: the header frame of the message
        :param method_frame: the method frame of the message
        :param channel: the channel we read the message from
        :param with_transaction: should we use a transaction

        :return: ReadResult object
        """

        delivery_tag = method_frame.delivery_tag
        assert delivery_tag is not None
        if with_transaction:
            transaction: InputTransaction = RabbitMQInputTransaction(cancellation_token=cancellation_token,
                                                                     device=self,
                                                                     channel=channel,
                                                                     delivery_tag=delivery_tag)
        else:
            transaction = NULLTransaction(self)
        headers: Dict[str, Any] = header_frame.headers or {}  # type: ignore
        # get the rabbitmq headers as for device headers
        rabbit_headers = self._get_rabbit_headers(method_frame, header_frame)

        buf = BytesIO(body)
        return ReadResult(message=Message(buf, headers),
                          device_headers=rabbit_headers,
                          transaction=transaction)

    def _get_frames_from_queue(self,
                               channel: 'BlockingChannel',
                               timeout: Optional[float],
                               with_transaction: bool) -> Tuple[
        Optional[bytes],
        Optional[spec.BasicProperties],
        Optional[Union[spec.Basic.Deliver, spec.Basic.GetOk]]
    ]:

        """
        gets the actual frame from queue. use_consumer effects this method

        :param BlockingChannel channel: the channel
        :param float timeout: the timeout
        :param bool with_transaction: do we operate within a transaction scope
        """
        # this could be auto_ack = not has_transaction, but it's less clear... so it's verbose here...
        if with_transaction:
            auto_ack = False
        else:
            auto_ack = True

        if self._use_consumer:
            if self._last_consumer_auto_ack is None:
                self._last_consumer_auto_ack = auto_ack

            if self._last_consumer_auto_ack != auto_ack:
                channel.cancel()
                self._last_consumer_auto_ack = auto_ack

            method_frame: Optional[Union[spec.Basic.Deliver, spec.Basic.GetOk]]
            header_frame: Optional[spec.BasicProperties]
            body: Optional[bytes]

            method_frame, header_frame, body = next(channel.consume(queue=self._queue_name,
                                                                    inactivity_timeout=timeout,
                                                                    arguments=self._consumer_args, auto_ack=auto_ack))
        else:
            method_frame, header_frame, body = channel.basic_get(queue=self._queue_name,  # type: ignore
                                                                 auto_ack=auto_ack)

        return body, header_frame, method_frame

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional['ReadResult']:
        """
        reads a stream from InputDevice (tries getting a message. if it fails, reconnects and tries again once)

        :param timeout: the timeout in seconds to block. negative number means no blocking
        :return: a tuple of stream and metadata from InputDevice, or (None, None) if no message is available
        """
        try:
            from pika.exceptions import AMQPConnectionError, AMQPChannelError
        except ImportError as exc:
            raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from exc

        try:
            return self._get_data_from_queue(cancellation_token=cancellation_token,
                                             timeout=timeout,
                                             with_transaction=with_transaction)
        except (AMQPConnectionError, AMQPChannelError):
            self._reconnect_device_manager()
            try:
                return self._get_data_from_queue(cancellation_token=cancellation_token,
                                                 timeout=timeout,
                                                 with_transaction=with_transaction)
            except Exception:
                self._logger.exception(f"AMQError thrown. failed to get message. device name: {self._queue_name}")
                raise
        except Exception as e:
            raise InputDeviceException('Error reading from device') from e

    def close(self):
        """
        closes the connection to device
        """
        try:
            if self._channel is not None and self._channel.is_open:
                if self._use_consumer:
                    self._channel.cancel()
                self._channel.close()
        except Exception:
            self._logger.warning('Error Closing Device', exc_info=True)

        self._channel = None


class RabbitMQInputDeviceManager(RabbitMQDeviceManagerBase, InputDeviceManager[RabbitMQInputDevice]):
    """
    rabbitmq input device manager
    """

    def __init__(self,
                 hosts: Union[List[str], str],
                 user: str,
                 password: str,
                 port: Optional[int] = None,
                 ssl_context: Optional[ssl.SSLContext] = None,
                 virtual_host: Optional[str] = None,
                 client_args: Optional[Dict[str, str]] = None,
                 heartbeat: int = 300,
                 connection_attempts: int = 5,
                 prefetch_count: int = 1,
                 use_consumer: bool = True,
                 blocked_connection_timeout: Optional[float] = None,
                 default_direct_exchange: Optional[str] = None
                 ):
        """
        This manager used to create RabbitMQ devices (direct queues)

        :param hosts: the hostname or a list of hostnames of the manager
        :param user: the username for the rabbitMQ manager
        :param password: the password for the rabbitMQ manager
        :param port: the port to connect the hosts to
        :param ssl_context: the ssl context to use. None means don't use ssl at all
        :param virtual_host: the virtual host to connect to
        :param client_args: the arguments to create the client with
        :param int heartbeat: heartbeat interval for the connection (between 0 and 65536
        :param int connection_attempts: Maximum number of retry attempts
        (-1 means not to handle poison messages at all, 0 means reject all redelivered messages right away)
        :param int prefetch_count: the number of unacked messages that can be consumed
        :param bool use_consumer: True to use the 'consume' method, False to use 'basic_get'
        :param blocked_connection_timeout: If not None,
            the value is a non-negative timeout, in seconds, for the
            connection to remain blocked (triggered by Connection.Blocked from
            broker); if the timeout expires before connection becomes unblocked,
            the connection will be torn down, triggering the adapter-specific
            mechanism for informing client app about the closed connection:
            passing `ConnectionBlockedTimeout` exception to on_close_callback
            in asynchronous adapters or raising it in `BlockingConnection`.

        :param default_direct_exchange: optional direct exchange to bind all the queues to (None means no bind)
        """
        super().__init__(hosts=hosts,
                         user=user,
                         password=password,
                         port=port,
                         ssl_context=ssl_context,
                         virtual_host=virtual_host,
                         client_args=client_args,
                         connection_type="Input",
                         heartbeat=heartbeat,
                         connection_attempts=connection_attempts,
                         blocked_connection_timeout=blocked_connection_timeout)

        self._prefetch_count = prefetch_count
        self._use_consumer = use_consumer
        self._default_direct_exchange = default_direct_exchange

    def _device_factory(self, device_name: str) -> RabbitMQInputDevice:

        return RabbitMQInputDevice(device_manager=self,
                                   queue_name=device_name,
                                   consumer_args=self._client_args,
                                   prefetch_count=self._prefetch_count,
                                   use_consumer=self._use_consumer)

    def get_input_device(self, name: str) -> RabbitMQInputDevice:
        """
        Returns an incoming device by name

        :param name: the name of the device to read from
        :return: an input device for 'device_name'
        """
        try:
            self.create_queue(queue_name=name,
                              passive=True,
                              direct_bind_to_exchange=self._default_direct_exchange)

            return self._device_factory(name)

        except Exception as e:
            message = f"Couldn't create input device '{name}'"
            self._logger.exception(message)
            raise InputDeviceException(message) from e

    def connect(self):
        """
        connects to the device manager
        """
        try:
            self._connect()
        except Exception as e:
            raise InputDeviceException('Could not connect to rabbitmq.') from e

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._disconnect()



import os
import shutil

import errno
import time

from messageflux.utils import KwargsException

MAX_LOCKFILE_AGE = 60  # max lockfile age in seconds


class AtomicMoveException(KwargsException):
    """
    an exception which is raised when atomic move fails
    """
    pass


def atomic_move(src: str, dest: str, lock_filename: str) -> bool:
    """
    moves a file atomically (hopefully) from src to dest, using lock_filename as lockfile

    :param src: the source file path
    :param dest: the destination file path
    :param lock_filename: the lockfile path
    :return: True if succeeded, False otherwise
    """
    fd = None
    filenames = [lock_filename]
    try:
        try:
            while True:
                if os.path.exists(lock_filename):
                    file_age = time.time() - os.stat(lock_filename).st_mtime
                    if file_age >= MAX_LOCKFILE_AGE:
                        lock_filename += '.new'
                        filenames.append(lock_filename)
                    else:
                        return False
                else:
                    break
            fd = os.open(lock_filename, os.O_RDWR | os.O_CREAT | os.O_EXCL)
        except (OSError, IOError):
            return False

        try:
            if not os.path.exists(src):
                return False
            shutil.move(src, dest)
            return True
        except (OSError, IOError) as e:
            if os.path.exists(dest):
                os.remove(dest)
            if e.errno == errno.ENOENT:  # no such file or directory
                return False
            raise AtomicMoveException(f"Error Moving file {src} to {dest}") from e
    finally:
        try:
            if fd is not None:
                os.close(fd)
        except Exception:
            pass
        for lock_filename in reversed(filenames):
            try:
                os.remove(lock_filename)
            except Exception:
                pass


def recursive_chmod(dir_name: str):
    """
    chmods to 0o777 the directory and its directory tree

    :param dir_name: the dir to chmod
    """
    os.chmod(dir_name, 0o777)
    for root, dirs, files in os.walk(dir_name):
        for d in dirs:
            try:
                os.chmod(os.path.join(root, d), 0o777)
            except OSError:
                pass
        for f in files:
            try:
                os.chmod(os.path.join(root, f), 0o777)
            except OSError:
                pass



from abc import ABCMeta, abstractmethod

from messageflux.iodevices.base import OutputDevice, OutputDeviceManager
from messageflux.iodevices.base.common import MessageBundle


class OutputTransformerBase(metaclass=ABCMeta):
    """
    transformer for output devices
    """

    def connect(self):
        """
        this method is called when the transformer device manager is connected
        """
        pass

    def disconnect(self):
        """
        this method is called when the transformer device manager is disconnected
        """
        pass

    @abstractmethod
    def transform_outgoing_message(self,
                                   output_device: 'TransformerOutputDevice',
                                   message_bundle: MessageBundle) -> MessageBundle:
        """
        Transform the message before it is sent to the underlying device.

        :param output_device: the output device that the transformer runs on
        :param message_bundle: the message bundle to transform
        :return: the transformed message bundle to send through the underlying device
        """
        pass


class TransformerOutputDevice(OutputDevice['TransformerOutputDeviceManager']):
    """
    a wrapper device that transforms outgoing messages
    """

    def __init__(self,
                 manager: 'TransformerOutputDeviceManager',
                 name: str,
                 inner_device: OutputDevice,
                 transformer: OutputTransformerBase):
        """

        :param manager: the device manager
        :param name: the name of the device
        :param inner_device: the inner device that this device wraps
        :param transformer: the output transformer to use
        """
        super().__init__(manager, name)
        self._transformer = transformer
        self._inner_device = inner_device

    def _send_message(self, message_bundle: MessageBundle):
        message_bundle = self._transformer.transform_outgoing_message(self, message_bundle)
        self._inner_device.send_message(message_bundle.message, message_bundle.device_headers)

    def close(self):
        """
        closes the inner device
        """
        self._inner_device.close()


class TransformerOutputDeviceManager(OutputDeviceManager[TransformerOutputDevice]):
    """
    a wrapper output device manager, that wraps the devices in transformer output devices
    """

    def __init__(self, inner_device_manager: OutputDeviceManager, transformer: OutputTransformerBase):
        """

        :param inner_device_manager: the inner device manager
        :param transformer: the transformer to use
        """
        self._inner_device_manager = inner_device_manager
        self._transformer = transformer

    def connect(self):
        """
        connects the inner device manager and the transformer
        """
        self._transformer.connect()
        self._inner_device_manager.connect()

    def disconnect(self):
        """
        disconnects the inner device manager and the transformer
        """
        self._inner_device_manager.disconnect()
        self._transformer.disconnect()

    def get_output_device(self, name: str) -> TransformerOutputDevice:
        """
        returns a wrapped output device
        :param name: the name of the device to get
        """
        inner_device = self._inner_device_manager.get_output_device(name)
        return TransformerOutputDevice(self, name, inner_device, self._transformer)



from .fs_poison_counter import FileSystemPoisonCounter
from .rabbitmq_input_device import RabbitMQInputDeviceManager, RabbitMQInputDevice, RabbitMQInputTransaction
from .rabbitmq_output_device import RabbitMQOutputDeviceManager, RabbitMQOutputDevice
from .rabbitmq_poison_counting_input_device import RabbitMQPoisonCountingInputDeviceManager, PoisonCounterBase



import threading
from abc import abstractmethod, ABCMeta
from time import time
from typing import List, Optional, Tuple, Union

from messageflux.iodevices.base import (InputTransactionScope,
                                        InputDeviceManager,
                                        AggregatedInputDevice,
                                        InputDevice, ReadResult)
from messageflux.server_loop_service import ServerLoopService


class MessageHandlingServiceBase(ServerLoopService, metaclass=ABCMeta):
    """
    a service thats reads from input devices and handles the messages
    """

    def __init__(self, *,
                 input_device_manager: InputDeviceManager,
                 input_device_names: Union[List[str], str],
                 use_transactions: bool = True,
                 read_timeout: float = 5,
                 max_batch_read_count: int = 1,
                 wait_for_batch_count: bool = False,
                 **kwargs):
        """

        :param input_device_manager: the device manager to read messages from
        :param input_device_names: the device names to read messages from
        :param use_transactions: whether to use transaction when reading from device
        :param read_timeout: the time (in seconds) to wait on device to return a message
        :param max_batch_read_count: the maximum batch size to read from device
        :param wait_for_batch_count: should the service wait the whole read_timeout for batch_count messages to be read.
        'False' means that it will read *up-to* batch_count messages, and process them immediately.
        'True' means that it will wait until read_timeout is passed, or batch_count messages has reached.
        :param **kwargs: passed to parent as is
        """
        super().__init__(**kwargs)
        self._input_device_manager = input_device_manager
        if isinstance(input_device_names, str):
            input_device_names = [input_device_names]
        self._input_device_names = input_device_names
        self._use_transactions = use_transactions
        self._read_timeout = max(read_timeout, 0)
        self._max_batch_read_count = max(max_batch_read_count, 1)
        self._wait_for_batch_count = wait_for_batch_count
        self._aggregate_input_device: Optional[AggregatedInputDevice] = None

    def _prepare_service(self):
        self._input_device_manager.connect()
        self._aggregate_input_device = self._input_device_manager.get_aggregate_device(self._input_device_names)

    def _server_loop(self, cancellation_token: threading.Event):
        assert self._aggregate_input_device is not None
        with InputTransactionScope(device=self._aggregate_input_device,
                                   with_transaction=self._use_transactions) as transaction_scope:
            batch: List[Tuple[InputDevice, ReadResult]] = []

            # read first message with _read_timeout anyway
            read_result = transaction_scope.read_message(cancellation_token=cancellation_token,
                                                         timeout=self._read_timeout)
            if read_result is not None:
                last_read_device = self._aggregate_input_device.last_read_device
                assert last_read_device is not None
                batch.append((last_read_device, read_result))

            end_time = time() + self._read_timeout
            for i in range(self._max_batch_read_count - 1):  # try to read the rest of the batch
                remaining_time = end_time - time()

                if remaining_time <= 0:
                    break

                if self._wait_for_batch_count:
                    timeout = remaining_time  # if wait_for_batch_count, try to read another message with remaining time
                else:
                    timeout = 0  # if not wait_for_batch, try to read another message without waiting at all

                read_result = transaction_scope.read_message(cancellation_token=cancellation_token,
                                                             timeout=timeout)
                if read_result is None:
                    break  # no more messages to read
                last_read_device = self._aggregate_input_device.last_read_device
                assert last_read_device is not None
                batch.append((last_read_device, read_result))

            if batch:
                self._handle_message_batch(batch)

            transaction_scope.commit()

    def _finalize_service(self, exception: Optional[Exception] = None):
        self._input_device_manager.disconnect()

    @abstractmethod
    def _handle_message_batch(self, batch: List[Tuple[InputDevice, ReadResult]]):
        """
        handles a batch of messages that was read from input devices
        :param batch: a list of tuples of input device and the ReadResult object that was read from it
        """
        pass


class BatchMessageHandlerBase(metaclass=ABCMeta):
    """
    a batch message handler base class. used to handle a batch of messages
    """

    def prepare(self):
        """
        called when the service starts.
        can be overrided by child class to perform some initialization logic
        """
        pass

    def shutdown(self):
        """
        called when the service stops.
        can be overrided by child class to perform some cleanup logic
        """
        pass

    @abstractmethod
    def handle_message_batch(self, batch: List[Tuple[InputDevice, ReadResult]]):
        """
        handles a batch of messages that was read from input devices
        :param batch: a list of tuples of input device and the ReadResult object that was read from it
        """
        pass


class BatchMessageHandlingService(MessageHandlingServiceBase):
    """
    a service that reads from input devices and handles the messages
    """

    def __init__(self, *,
                 batch_handler: BatchMessageHandlerBase,
                 **kwargs):
        """

        :param batch_handler: the message handler to use.
        :param **kwargs: passed to parent as is
        """
        super().__init__(**kwargs)
        self._message_handler = batch_handler

    def _prepare_service(self):
        super()._prepare_service()
        self._message_handler.prepare()

    def _finalize_service(self, exception: Optional[Exception] = None):
        self._message_handler.shutdown()
        super()._finalize_service(exception)

    def _handle_message_batch(self, batch: List[Tuple[InputDevice, ReadResult]]):
        self._message_handler.handle_message_batch(batch)


class MessageHandlerBase(metaclass=ABCMeta):
    """
    a message handler base class. used to handle a single message
    """

    def prepare(self):
        """
        called when the service starts.
        can be overrided by child class to perform some initialization logic
        """
        pass

    def shutdown(self):
        """
        called when the service stops.
        can be overrided by child class to perform some cleanup logic
        """
        pass

    @abstractmethod
    def handle_message(self, input_device: InputDevice, read_result: ReadResult):
        """
        handles a single message from the device

        :param input_device: the input device that the message was read from
        :param read_result: the read result returned from the device
        """
        pass


class MessageHandlingService(BatchMessageHandlingService):
    """
    a service that reads a SINGLE message from devices and handles it
    """

    class _BatchMessageHandlerAdapter(BatchMessageHandlerBase):
        def __init__(self, massage_handler: MessageHandlerBase):
            self._message_handler = massage_handler

        def prepare(self):
            """
            called when the service starts.
            can be overrided by child class to perform some initialization logic
            """
            self._message_handler.prepare()

        def shutdown(self):
            """
            called when the service stops.
            can be overrided by child class to perform some cleanup logic
            """
            self._message_handler.shutdown()

        def handle_message_batch(self, batch: List[Tuple[InputDevice, ReadResult]]):
            """
            handles a batch of messages that was read from input devices
            :param batch: a list of tuples of input device and the ReadResult object that was read from it
            """
            for input_device, read_result in batch:
                self._message_handler.handle_message(input_device=input_device,
                                                     read_result=read_result)

    def __init__(self, *, message_handler: MessageHandlerBase, **kwargs):
        """
        :param **kwargs: passed to parent as is
        """
        super().__init__(batch_handler=self._BatchMessageHandlerAdapter(message_handler), **kwargs)



from enum import Enum

LOG_TYPE_EXTRA = 'log_type'


class LogType(Enum):
    """
    this enum defines log_types
    """
    INPUT_LOG_TYPE = 'input_log'
    OUTPUT_LOG_TYPE = 'output_log'
    FINISH_LOG_TYPE = 'finish_log'



import logging

import sys

DEFAULT_LOG_FORMATTER = logging.Formatter(
    "%(asctime)-15s [%(levelname)s] [PID:%(process)d] %(message)s")
DEFAULT_LOG_HANDLER = logging.StreamHandler(sys.stdout)
DEFAULT_LOG_HANDLER.setFormatter(DEFAULT_LOG_FORMATTER)
